/*******************************************************************************
 * Quick access to the Components object.
 ******************************************************************************/
const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;
/*******************************************************************************
 * Component definitions.
 ******************************************************************************/
const ITEM_CLASSNAME = "Netscape News Live Item";
const ITEM_CLASSID = Components.ID("{71c420fe-c793-0842-a340-f1af29a6cc8a}");
const ITEM_CONTRACTID = "@netscape.com/newslive/item;1";

const SERVICE_CLASSNAME = "Netscape News Live Service";
const SERVICE_CLASSID = Components.ID("{99b04049-80f1-1544-b311-be22873404c3}");
const SERVICE_CONTRACTID = "@netscape.com/newslive/service;1";
/*******************************************************************************
 * Contract IDs.
 ******************************************************************************/
const BUNDLE_CONTRACTID = "@mozilla.org/intl/stringbundle;1";
const PREF_CONTRACTID = "@mozilla.org/preferences-service;1";
const OB_CONTRACTID = "@mozilla.org/observer-service;1";
const REQUEST_CONTRACTID = "@mozilla.org/xmlextras/xmlhttprequest;1";
const TIMER_CONTRACTID = "@mozilla.org/timer;1";
const STORAGE_CONTRACTID = "@mozilla.org/storage/service;1"; 
const WRAPPER_CONTRACTID = "@mozilla.org/storage/statement-wrapper;1";
const ARRAY_CONTRACTID = "@mozilla.org/array;1";
const DIR_CONTRACTID = "@mozilla.org/file/directory_service;1";
const IO_CONTRACTID = "@mozilla.org/network/io-service;1";
const FPR_CONTRACTID = "@mozilla.org/feed-processor;1";
const UNESCAPE_CONTRACTID = "@mozilla.org/feed-unescapehtml;1";
const CONSOLE_CONTRACTID = "@mozilla.org/consoleservice;1";
const ERROR_CONTRACTID = "@mozilla.org/scripterror;1";
/*******************************************************************************
 * Observer topics
 ******************************************************************************/
const TOPIC_TIMER = "timer-callback";
const TOPIC_NEWSLIVE = "netscape-newslive";
const TOPIC_SHUTDOWN_APP = "quit-application";
const TOPIC_SHUTDOWN_XPCOM = "xpcom-shutdown";
const TOPIC_PREF = "nsPref:changed";
/*******************************************************************************
 * Default values used across components.
 ******************************************************************************/
const DEFAULT_BRANCH = "extensions.netscape.newslive.";
const DEFAULT_BUNDLE = "chrome://browser/locale/netscape-newslive/newslive.properties";
const DEFAULT_FAILED = "newslive.loading.failed";
const DEFAULT_SEC = 1000;                                    // 1 second
const DEFAULT_MIN = 60*DEFAULT_SEC;                          // 1 minute
const DEFAULT_HOUR = 60*DEFAULT_MIN;                         // 1 hour
const DEFAULT_DAY = 24*DEFAULT_HOUR;                         // 1 day
const DEFAULT_ERROR = "http://www.mozilla.org/newlayout/xml/parsererror.xml";
const DEFAULT_ROOT = "newslive_root";
/*******************************************************************************
 * Status checker values.
 ******************************************************************************/
const CHECKER_URL = "endpoint.api";
const CHECKER_PAUSED = "Server error: " + SERVICE_CLASSNAME + " temporarily disabled for $1 minutes.";
/*******************************************************************************
 * API Checker values
 ******************************************************************************/
const API_EXPIRES = "feedsExpire";
const API_DEFAULTS = "feedsDefault";
const API_MAX = 60*DEFAULT_DAY;                              // 60 days
const API_SAME = "sameFeeds";
const API_NEW = "newFeeds";
const API_CHANGED = "changedFeeds";
/*******************************************************************************
 * Feeds Checker values
 ******************************************************************************/
const FEEDS_COUNT = "itemCount";
/*******************************************************************************
 * NewsliveService values.
 ******************************************************************************/
const SERVICE_DIR = "ProfD";
const SERVICE_FILE = "newslive.sqlite";
const SERVICE_SCHEMA = 2;
const SERVICE_TABLES = {
  items: "id TEXT PRIMARY KEY, parentId TEXT, title TEXT, URL TEXT, " +
         "iconURL TEXT, refreshRate INTEGER, subscribed BOOLEAN, " +
         "lastUpdated REAL, sortIndex INTEGER",
  schema: "version INTEGER PRIMARY KEY"
};
const SERVICE_STMTS = {
  selectItems: "SELECT * FROM items ORDER BY sortIndex",
  insertItem: "INSERT INTO items (id, parentId, title, URL, iconURL, " +
              "refreshRate, subscribed, lastUpdated, sortIndex) VALUES " +
              "(:id, :parentId, :title, :URL, :iconURL, :refreshRate, " +
              ":subscribed, :lastUpdated, :sortIndex)",
  updateItem: "UPDATE items set parentId = :parentId, title = :title, " +
              "URL = :URL, iconURL = :iconURL, refreshRate = :refreshRate, " +
              "subscribed = :subscribed, lastUpdated = :lastUpdated, " +
              "sortIndex = :sortIndex WHERE id = :id",              
  deleteItem: "DELETE FROM items WHERE id = :id",    
  deleteItems: "DELETE FROM items WHERE parentId = :parentId"
};
const SERVICE_TIMER = 30*DEFAULT_SEC;
const SERVICE_COMPACT = "VACUUM";
const SERVICE_DISABLED = "disabled";
const SERVICE_ADD_ITEM = "addItem";
const SERVICE_REMOVE_ITEM = "removeItem";
const SERVICE_UPDATE_ITEM = "updateItem";
const SERVICE_ADD_ITEMS = "addItems";
const SERVICE_REMOVE_ITEMS = "removeItems";
const SERVICE_STATE = "changedState";
const SERVICE_URL = "about:newslive";
/*******************************************************************************
 * Helper function to create a mozIStorage statement wrappter that can be used 
 * to access the database.
 *
 * @param     conn           The database connection that the statement 
 *                           is applied to.
 * @param     sql            The sql statement that should be created.
 * @return                   The storage wrapper created.
 ******************************************************************************/
function createStatement(conn, sql) {
  var statement = conn.createStatement(sql);
  var rv = Cc[WRAPPER_CONTRACTID].
           createInstance(Ci.mozIStorageStatementWrapper);
  rv.initialize(statement);
  
  return rv;
}
/*******************************************************************************
 * Helper function to create an item component from a mozIStorage row.
 *
 * @param     row            The row containing the item data.
 * @return                   The created item.
 ******************************************************************************/
function createItemFromRow(row) {
  var rv = new NewsliveItem();
  rv.init(row.id, row.parentId, row.title, row.URL, row.iconURL, row.refreshRate, 
          row.subscribed, row.lastUpdated, row.sortIndex);
  
  return rv;
}
/*******************************************************************************
 * Helper function to sort objects by sortIndex.
 ******************************************************************************/
function sorter(a, b) {
  if (a.sortIndex > b.sortIndex)
    return 1;
  else
    return -1;
}
/*******************************************************************************
 * Function to log a message to the error console
 *
 * @param     message        Message to log
 ******************************************************************************/
function logMessage(message) {
  
  // create an error object
  var scriptError = Cc[ERROR_CONTRACTID].createInstance(Ci.nsIScriptError);
  var sourceName = __LOCATION__.path;
  var flag = Ci.nsIScriptError.warningFlag;
  scriptError.init(message, sourceName, "", 0, 0, flag, SERVICE_CONTRACTID);
  
  // log to console
  var consoleService = Cc[CONSOLE_CONTRACTID].getService(Ci.nsIConsoleService);
  consoleService.logMessage(scriptError);
}
/*******************************************************************************
 * Function to get a string from the bundle
 *
 * @param     name           Name of the string to get.
 * @return                   The string or null if string not found.
 ******************************************************************************/
function getString(name) {
  var sbService = Cc[BUNDLE_CONTRACTID].getService(Ci.nsIStringBundleService);
  var bundle = sbService.createBundle(DEFAULT_BUNDLE);
  if (bundle)
    return bundle.GetStringFromName(name);
  return null;
}
/*******************************************************************************
 * Object used to check with the feeds api or rss feed for newslive data.
 *
 * @param     service        The newslive service.
 ******************************************************************************/
function StatusChecker(service) {
  this.service = service;
  this.load();
}
StatusChecker.prototype = {

  _service: null,
  get service() { return this._service; },
  set service(val) { this._service = val; },
  
  _requests: null,
  get requests() { return this._requests; },
  set requests(val) { this._requests = val; },
      
  _errorCount: null,
  get errorCount() { return this._errorCount; },
  set errorCount(val) { this._errorCount = val; },
   
  _errorTimer: null,
  get errorTimer() { return this._errorTimer; },
  set errorTimer(val) { this._errorTimer = val; },
    
  // nsISupports
  QueryInterface: function sc_QI(iid) {
    if (iid.equals(Ci.nsIObserver) ||
        iid.equals(Ci.nsISupports))
      return this;
    throw Cr.NS_ERROR_NO_INTERFACE; 
  },

  // nsIObserver
  observe: function sc_observe(subject, topic, data) {
  
    // do nothing if not timer topic
    if (topic != TOPIC_TIMER)
      return;
      
    // re-enable after timer
    this.service._paused = false;
  },
  
  load: function sc_load() {
    this.requests = {};
    this.errorCount = 0;
    this.errorTimer = Cc[TIMER_CONTRACTID].createInstance(Ci.nsITimer); 
  },
  
  unload: function sc_unload() {
    
    // stop the timer
    this.errorTimer.cancel();
    
    // stop any requests
    var requests = this.requests;
    for (var id in requests)
      requests[id].abort();
      
    // remove variables
    this.service = null;
    this.requests = null;  
    this.errorCount = null;
    this.errorTimer = null;
  },
  
  reloadItems: function sc_reloadItems(parentId) {
  
    // reload feed api
    var item; 
    if (!parentId) {
      if (this.needsUpdated(null, true))
        this.requestUpdate(null);
    
    // reload all feeds
    } else if (parentId == DEFAULT_ROOT) {
      var items = this.service.getItems(DEFAULT_ROOT, true);
      var index = items.length;
      while (index--) {
        item = items.queryElementAt(index, Ci.nsINewsliveItem);
        if (this.needsUpdated(item, true))
          this.requestUpdate(item);
      }
      
    // reload specific feed
    } else {
      item = this.service.getItem(parentId);
      if (this.needsUpdated(item, true))
        this.requestUpdate(item);
    }
  },
  
  requestUpdates: function sc_requestUpdates() {
  
    // request feed api update
    if (this.needsUpdated(null, false))
      this.requestUpdate(null);
      
    // request update of feeds  
    else {
      var items = this.service.getItems(DEFAULT_ROOT, true);
      for (var index=0; index<items.length; index++) {
        var item = items.queryElementAt(index, Ci.nsINewsliveItem);
        if (this.needsUpdated(item, false))
          this.requestUpdate(item);
      }
    }
  },
  
  needsUpdated: function sc_needsUpdated(item, force) {
    var id = (!item) ? "api" : item.id;
    
    // currently being update we do not need another
    if (this.requests.hasOwnProperty(id))
      return false;
    
    // force so we update
    if (force)
      return true;
    
    // api expiration checking
    var now = (new Date()).getTime();
    var expires;
    if (id == "api") {  

      // api expired we need an update    
      expires = Number(this.service._branch.getCharPref(API_EXPIRES));
      expires = (new Date(expires)).getTime();
      if (expires <= now)
        return true;
    
      // service has no feeds we need an update  
      if (!this.service.hasItems(DEFAULT_ROOT, false))
        return true;
    
    // rss expiration checking  
    } else {
    
      // feed expired needs updated   
      expires = item.lastUpdated + (item.refreshRate*DEFAULT_SEC);
      if (expires <= now)
        return true;
    
      // no items so needs updated
      if (!this.service.hasItems(id, false))
        return true;  
    }
      
    // does not need updated
    return false;    
  },
    
  requestUpdate: function sc_requestUpdate(item) {
    var id = (!item) ? "api" : item.id;
    
    // remove the children
    if (id != "api")
      this.service.removeItems(id);
               
    // create an XML HTTP request  
    var self = this;
    var request = Cc[REQUEST_CONTRACTID].createInstance(Ci.nsIXMLHttpRequest);
    var api = (id == "api") ? this.service._branch.getCharPref(CHECKER_URL) : item.URL; 
    request.open("GET", api, true);
    request.overrideMimeType("text/xml");
    request.setRequestHeader("X-Netscape", "Navigator");        
    request.QueryInterface(Ci.nsIJSXMLHttpRequest).onerror = function() { self.onError(item); };
    request.QueryInterface(Ci.nsIJSXMLHttpRequest).onload = function() { self.onLoad(item); };
    this.requests[id] = request;    
      
    // send the request
    this.requests[id].send(null); 
  },
  
  onLoad: function sc_onLoad(item) {
      
    // parse the data
    var id = (!item) ? "api" : item.id;
    var request = this.requests[id];
    var parser = (id == "api") ? new APIParser(this) : new FeedParser(item, this);
    parser.load(request);
  },
  
  onError: function sc_onError(item) {
      
    // release the request 
    var id = (!item) ? "api" : item.id;
    delete this.requests[id];
    
    // already paused
    if (this.service.paused)
      return;
          
    // log to the console
    this.errorCount++;
    var delay = Math.pow(this.errorCount, 2);
    var message = CHECKER_PAUSED.replace("$1", delay);
    logMessage(message);
        
    // increment the error counter and start the disabled timer   
    this.service._paused = true;
    this.errorTimer.init(this, delay*DEFAULT_MIN, Ci.nsITimer.TYPE_ONE_SHOT); 
  } 
};
/*******************************************************************************
 * Object used to parse the api data.
 *
 * @param     checker        The status checker.
 ******************************************************************************/
function APIParser(checker) {
  this.checker = checker;
}
APIParser.prototype = {

  _checker: null,
  get checker() { return this._checker; },
  set checker(val) { this._checker = val; },
    
  // nsISupports
  QueryInterface: function api_QI(iid) {
    if (iid.equals(Ci.nsISupports))
      return this;
    throw Cr.NS_ERROR_NO_INTERFACE; 
  },
   
  load: function api_load(request) {
    var service = this.checker.service;
          
    // get the feed items from the xml
    var items = null;
    try {
      items = this.createItems(request.responseXML);
    } catch(e) {}
    
    // data could not be created
    if (!items || items.length === 0) {
      this.checker.onError(null);
      this.checker = null;
      return;
    }
     
    // set a new expires date
    var now = (new Date()).getTime();
    var max = now + API_MAX;
    var expires;
    try {
      expires = request.getResponseHeader("Expires");
      expires = Date.parse(expires);
      if (expires < now || expires > max)
        expires = max;
    } catch(e) {
      expires = max;
    }   
    service._branch.setCharPref(API_EXPIRES, String(expires));
        
    // set the data notify the front-end
    var state = this.setItems(items);
    service._notify(null, state);
    
    // request an update if the state changed
    if (state != API_SAME)
      this.checker.requestUpdates();
         
    // cleanup
    this.checker.errorCount = 0;
    delete this.checker.requests["api"];
    this.checker = null;    
  },
  
  createItems: function api_createItems(xml) {
    var service = this.checker.service;
      
    // parse error
    if (!xml || xml.documentElement.namespaceURI == DEFAULT_ERROR)
      return null; 

    // setup the needed variables
    var ueService = Cc[UNESCAPE_CONTRACTID].getService(Ci.nsIScriptableUnescapeHTML);   
    var items = Cc[ARRAY_CONTRACTID].createInstance(Ci.nsIMutableArray); 
    var nodes = xml.getElementsByTagName("feed");
    var subscriptions = service._branch.getCharPref(API_DEFAULTS).split(",");    
    var defaultIndex = subscriptions.length -1;     
    var lastUpdated = 0;           
        
    // loop through nodes and create items
    for (var i=0; i<nodes.length; i++) {
      var node = nodes[i];
      
      // get the item values
      var id = node.getAttribute("id");
      var title = node.getElementsByTagName("name")[0].textContent;
      title = ueService.unescape(title);
      var URL = node.getElementsByTagName("url")[0].textContent;
      var iconURL = node.getElementsByTagName("icon")[0].textContent;
      var refreshRate = Number(node.getAttribute("refreshrate"));
      var index = subscriptions.indexOf(id);
      var subscribed = (index != -1);
      
      // determine the sortIndex
      ++defaultIndex;
      var sortIndex = (subscribed) ? index : defaultIndex;
      sortIndex = sortIndex*100;
      
      // create the item
      var item = new NewsliveItem();
      item.init(id, DEFAULT_ROOT, title, URL, iconURL, refreshRate, subscribed, 
                lastUpdated, sortIndex);
                
      // append to the returning array
      items.appendElement(item, false);
    }
    
    return items.QueryInterface(Ci.nsIArray);           
  },
  
  setItems: function api_setItems(items, request) {
    var service = this.checker.service;
      
    // set the notification state
    var rv = API_SAME;
    
    // no feeds in service so add
    if (!service.hasItems(DEFAULT_ROOT, false)) {
      rv = API_NEW;
      service.addItems(items);
    
    // merge with existing feeds    
    } else if (this.mergeItems(items))
      rv = API_CHANGED;

    return rv;
  },
  
  mergeItems: function api_mergeItems(items) {
    var service = this.checker.service;   
    var rv = false;
    
    // function to merge the new and old items into one list
    var allItems = [];
    function merge(newItems, newValue) {
      var index = newItems.length;
      while (index--) {
        var item = { "flag": newValue, "value": newItems.queryElementAt(index, Ci.nsINewsliveItem) };
        allItems[allItems.length] = item;
      }  
    }

    // merge into a single list
    var oldItems = service.getItems(DEFAULT_ROOT, false);    
    merge(items, true);
    merge(oldItems, false);
    
    // function to sort by item id then value
    function sortById(a, b) {
      if (a.value.id > b.value.id)
        return 1;
      else if (a.value.id == b.value.id) {
        if (a.flag == true)
          return -1;
        else
          return 1;
      } else
        return -1;
    }
    
    // sort items
    allItems.sort(sortById);
    
    // loop through all items
    for (var i=0; i<allItems.length; i++) {
      var item1 = allItems[i];
      var item2 = allItems[i+1];
      
      // combine items since they match
      if (item1.value.id == item2.value.id) {
        var newItem = item1.value;
        var oldItem = item2.value;
        oldItem.init(oldItem.id, newItem.parentId, newItem.title, newItem.URL, 
                     newItem.iconURL, newItem.refreshRate, oldItem.subscribed, 
                     oldItem.lastUpdated, oldItem.sortIndex);
        service.updateItem(oldItem);   
        i++;
            
      // new item so add
      } else if (item1.flag == true){
        service.addItem(item1.value);
        rv = true;
        
      // old item so remove
      } else {
        service.removeItems(item1.value.id);
        service.removeItem(item1.value.id);
        rv = true;
      }
    }
         
    return rv;
  }
};
/*******************************************************************************
 * Object used to parse the rss data.
 *
 * @param      item          The item being parsed.
 * @param     checker        The status checker.
 ******************************************************************************/
function FeedParser(item, checker) {
  this.item = item;
  this.checker = checker;
}
FeedParser.prototype = {

  _item: null,
  get item() { return this._item; },
  set item(val) { this._item = val; },
  
  _checker: null,
  get checker() { return this._checker; },
  set checker(val) { this._checker = val; },
  
  _processor: null,
  get processor() { return this._processor; },
  set processor(val) { this._processor = val; },
    
  // nsISupports
  QueryInterface: function sc_QI(iid) {
    if (iid.equals(Ci.nsIFeedResultListener) ||
        iid.equals(Ci.nsISupports))
      return this;
    throw Cr.NS_ERROR_NO_INTERFACE; 
  },
    
  load: function feed_load(request) {
    
    // get the feed processor
    this.processor = Cc[FPR_CONTRACTID].createInstance(Ci.nsIFeedProcessor);
    this.processor.listener = this;
            
    // no text returned
    var data = request.responseText;
    if (!data || data.length === 0) {
      this.onError();
      return;
    }
    
    // hand the data off to the feed processor
    var ioService = Cc[IO_CONTRACTID].getService(Ci.nsIIOService);
    var uri = ioService.newURI(this.item.URL, null, null);
    try {
      this.processor.parseFromString(data, uri);
    } catch(e) {
      this.onError();
    }  
  },
      
  handleResult: function feed_handleResult(result) {
    var service = this.checker.service;
    var item = this.item;
    try {
      
      // invalid feed just return
      if (!result || !result.doc || result.bozo)
        throw "invalid feed";
      
      // loop through the feed entries    
      var title, url, entry;
      var items = Cc[ARRAY_CONTRACTID].createInstance(Ci.nsIMutableArray);
      var feed = result.doc.QueryInterface(Ci.nsIFeed);
      var itemCount = service._branch.getIntPref(FEEDS_COUNT);
      for (var i=0; i< feed.items.length; i++) {
      
        // max items hit leave loop
        if (i >= itemCount)
          break;
        
        // get the entry title  
        entry = feed.items.queryElementAt(i, Ci.nsIFeedEntry);
        if (entry.title)
          title = entry.title.plainText();
        else if (entry.updated)
          title = entry.updated;
         
        // get the entry uri
        if (entry.link)
          uri = entry.link;

        // create the item
        if (uri && title)
          items.appendElement(this.createItem(uri, title, i*100), false);    
      }
      
      // add the items to the service
      if (items.length > 0)
        service.addItems(items.QueryInterface(Ci.nsIArray));

      // no errors
      this.updateFeed();
      this.checker.errorCount = 0;
      delete this.checker.requests[item.id];
      this.item = null;
      this.checker = null;
      this.processor.listener = null;
      this.processor = null;
      
    // had error  
    } catch(e) {
      this.onError();
    }
  },


  insertFailed: function ic_insertFailed() {
    var service = this.checker.service;
    var id = this.item.id + "_failed";
    var title = getString(DEFAULT_FAILED);
    var parentId = this.item.id;
    var now = (new Date()).getTime();      
    
    // insert the failed item
    var item = new NewsliveItem();
    item.init(id, parentId, title, SERVICE_URL, "", 0, false, now, 0); 
    service.addItem(item);
  },
      
  onError: function feed_onError() {
    this.insertFailed();
    this.checker.onError(this.item);
    this.item = null;
    this.checker = null;
    this.processor.listener = null;
    this.processor = null;
  },
    
  updateFeed: function feed_updateFeed() {
    var now = (new Date()).getTime();      
    this.item.lastUpdated = now;
    this.checker.service.updateItem(this.item);
  },
  
  createItem: function feed_createItemItem(uri, title, index) {
    var feed = this.item;
    var item = new NewsliveItem();
    var id = feed.id + "_" + index;
    var now = (new Date()).getTime();          
    item.init(id, feed.id, title, uri.spec, "", feed.refreshRate, false, now, index);
    
    return item;
  }  
};
/*******************************************************************************
 * Interface used to describe a news live feed or feed item.  This interface is 
 * used to represent meta data describing a netscape news feed.
 * 
 * @version   1.0
 ******************************************************************************/
function NewsliveItem() {}
NewsliveItem.prototype = {
  _id: null,  
  _parentId: null,
  _title: null,
  _URL: null,
  _iconURL: null,
  _refreshRate: null,
  _subscribed: null,
  _lastUpdated: null,
  _sortIndex: null,
  
  // nsISupports
  QueryInterface: function nli_QI(iid) {
    if (iid.equals(Ci.nsINewsliveItem) ||
        iid.equals(Ci.nsISupports))
      return this;
    throw Cr.NS_ERROR_NO_INTERFACE; 
  },
    
  // nsINewsliveItem
  get id() { return this._id; },
  get parentId() { return this._parentId; },
  get title() { return this._title; },
  get URL() { return this._URL; },
  get iconURL() { return this._iconURL; },
  get refreshRate() { return this._refreshRate; },
  get subscribed() { return this._subscribed; },
  get lastUpdated() { return this._lastUpdated; },
  get sortIndex() { return this._sortIndex; },
  
  set subscribed(val) { this._subscribed = val; },
  set lastUpdated(val) { this._lastUpdated = val; },
  set sortIndex(val) { this._sortIndex = val; },
  
  init: function nli_init(id, parentId, title, URL, iconURL, refreshRate, 
                          subscribed, lastUpdated, sortIndex) {
    this._id = id;
    this._parentId = parentId;
    this._title = title;
    this._URL = URL;
    this._iconURL = iconURL;
    this._refreshRate = refreshRate;
    this._subscribed = subscribed;
    this._lastUpdated = lastUpdated;
    this._sortIndex = sortIndex;
  },

  clone: function nli_clone() {
    var rv = new NewsliveItem();
    rv.init(this._id, this._parentId, this._title, this._URL, this._iconURL,
            this._refreshRate, this._subscribed, this._lastUpdated,  
            this._sortIndex);
    
    return rv;
  }          
};
/*******************************************************************************
 * The main news live interface.  This interface takes care of updating the 
 * list of feeds and the individual feed items.  It notifies the front-end
 * when updates take place.
 * 
 * @version   1.0
 ******************************************************************************/
function NewsliveService() {
  this._load();
}
NewsliveService.prototype = {
  _obService: null,
  _branch: null,
  _disabled: null,  
  _paused: null,  
  _items: null,
  _conn: null,
  _statements: null,  
  _checker: null,
  _timer: null,
  
  // nsISupports
  QueryInterface: function nls_QI(iid) {
    if (iid.equals(Ci.nsINewsliveService) ||
        iid.equals(Ci.nsIObserver) ||
        iid.equals(Ci.nsISupports))
      return this;
    throw Cr.NS_ERROR_NO_INTERFACE; 
  },
  
  // nsIObserver
  observe: function nls_observe(subject, topic, data) {
    switch (topic) {
    
    case TOPIC_SHUTDOWN_APP:
      this._unload();
      break;
              
    case TOPIC_SHUTDOWN_XPCOM:
      this._unloadFinal();
      break;
    
    case TOPIC_TIMER:
      this._tryCheck();
      break;
    
    case TOPIC_PREF:
      this._prefUpdate(data);
      break;
          
    default:
      return;
    }
  },
  
  // nsINewsliveService
  get disabled() { return this._disabled; },
  get paused() { return this._paused; },
     
  get databaseFile() { 
    if (!this._conn) 
      return null; 
    return this._conn.databaseFile; 
  },
  
  compactDB: function nls_compactDB() {
    if (this._conn.transactionInProgress)
      throw Cr.NS_ERROR_FAILURE;    
    this._conn.executeSimpleSQL(SERVICE_COMPACT);  
  },
  
  hasItem: function nls_hasItem(id) { 
    return this._items.some(function(o) { return o.id == id; });
  },
  
  hasItems: function nls_hasItems(parentId, includeFailed) { 
    return this._items.some(function(o) { return ((o.parentId == parentId) && (includeFailed || (o.URL != SERVICE_URL))) });
  },
    
  getItem: function nls_getItem(id) {
  
    // do not have the item
    if (!this.hasItem(id))
      return null;  
      
    // return the first matching item
    return this._items.filter(function(o) { return o.id == id; }).shift();
  },
    
  getItems: function nls_getItems(parentId, subscribed) { 
    var rv = Cc[ARRAY_CONTRACTID].createInstance(Ci.nsIMutableArray);      
    
    // function to check first parentId then subscribed if root
    function check(o) {
      if (parentId != DEFAULT_ROOT)
        return (parentId == o.parentId);
      else {
        if (parentId == o.parentId) {
          if (!subscribed || (subscribed && o.subscribed))
            return true;
          else
            return false;
        }
        return false;
      }
    }
    
    // filter and sort the items
    var items = this._items.filter(check);
    items.sort(sorter);
    
    // loop through and add to the array
    var index = 0;
    while (index < items.length) {
      rv.appendElement(items[index], false);
      index++;
    }
    
    // return the array
    return rv.QueryInterface(Ci.nsIArray);
  },
  
  addItem: function nls_addItem(item) { 
  
    // have the item so update
    if (this.hasItem(item.id)) {
      this.updateItem(item);
      return;
    }

    // save the item in memory
    this._items[this._items.length] = item.clone();
    
    // write to the database
    this._conn.beginTransaction();
    var statement = this._statements["insertItem"];
    var params = statement.params;
    params.id = item.id;
    params.parentId = item.parentId;
    params.title = item.title;
    params.URL = item.URL;
    params.iconURL = item.iconURL;
    params.refreshRate = item.refreshRate;
    params.subscribed = item.subscribed;
    params.lastUpdated = item.lastUpdated;
    params.sortIndex = item.sortIndex;
    statement.execute();
    this._conn.commitTransaction();
    
    // notify the front-end
    this._notify(item, SERVICE_ADD_ITEM);
  },   
    
  addItems: function nls_addItems(items) {

    // remove items from the parent
    var item = items.queryElementAt(0, Ci.nsINewsliveItem);
    if (this.hasItems(item.parentId, true))
      this.removeItems(item.parentId);
      
    // loop through the array
    var index = items.length;
    var statement, params;
    this._conn.beginTransaction();         
    while (index--) {
      item = items.queryElementAt(index, Ci.nsINewsliveItem);   

      // save to memory
      this._items.push(item.clone());
      
      // write to database
      statement = this._statements["insertItem"];
      params = statement.params;
      params.id = item.id;
      params.parentId = item.parentId;
      params.title = item.title;
      params.URL = item.URL;
      params.iconURL = item.iconURL;
      params.refreshRate = item.refreshRate;
      params.subscribed = item.subscribed;
      params.lastUpdated = item.lastUpdated;
      params.sortIndex = item.sortIndex;
      statement.execute();
    }
    this._conn.commitTransaction(); 
    
    // notify the front-end
    this._notify(items, SERVICE_ADD_ITEMS);
  },
  
  updateItem: function nls_updateItem(item) { 
  
    // do not have the item so return
    if (!this.hasItem(item.id))
      return;
  
    // save the item to memory
    var items = this._items.filter(function(o) { return o.id != item.id; });
    items[items.length] = item.clone();
    this._items = items;
    
    // write to database
    this._conn.beginTransaction();
    var statement = this._statements["updateItem"];
    var params = statement.params;
    params.id = item.id;
    params.parentId = item.parentId;
    params.title = item.title;
    params.URL = item.URL;
    params.iconURL = item.iconURL;
    params.refreshRate = item.refreshRate;
    params.subscribed = item.subscribed;
    params.lastUpdated = item.lastUpdated;
    params.sortIndex = item.sortIndex;
    statement.execute();
    this._conn.commitTransaction();

    // notify the front-end
    this._notify(item, SERVICE_UPDATE_ITEM);
  },
       
  removeItem: function nls_removeItem(id) {
  
    // do not have the item so return
    var item = this.getItem(id);
    if (!item)
      return;

    // remove from memory
    var items = this._items.filter(function(o) { return o.id != item.id; });
    this._items = items; 
    
    // remove from database
    this._conn.beginTransaction();
    var statement = this._statements["deleteItem"];
    var params = statement.params;
    params.id = item.id;
    statement.execute();
    this._conn.commitTransaction();
    
    // notify the front-end
    this._notify(item, SERVICE_REMOVE_ITEM);      
  },
  
  removeItems: function nls_removeItems(parentId) { 
  
    // do not have the parent
    var item = this.getItem(parentId);
    if (parentId != DEFAULT_ROOT && !item)
      return;
    
    //  remove from memory
    var items = this._items.filter(function(o) { return o.parentId != parentId; });
    this._items = items; 

    // remove from database
    this._conn.beginTransaction();
    var statement = this._statements["deleteItems"];
    var params = statement.params;
    params.parentId = parentId;
    statement.execute();
    this._conn.commitTransaction();
    
    // notify the front-end
    this._notify(item, SERVICE_REMOVE_ITEMS);      
  },
  
  reloadItems: function nls_reloadItems(parentId) {
    this._checker.reloadItems(parentId);
  },
        
  _load: function nls__load() {
    
    // get the pref branch and observe it
    this._branch = Cc[PREF_CONTRACTID].getService(Ci.nsIPrefService).
                   getBranch(DEFAULT_BRANCH);
    this._branch.QueryInterface(Ci.nsIPrefBranch2).addObserver("", this, false);
    
    // get the observer service and add ourself as an observer                     
    this._obService = Cc[OB_CONTRACTID].getService(Ci.nsIObserverService);
    this._obService.addObserver(this, TOPIC_SHUTDOWN_APP, false);
    this._obService.addObserver(this, TOPIC_SHUTDOWN_XPCOM, false);
    
    // setup variables
    this._items = [];
    this._disabled = this._branch.getBoolPref(SERVICE_DISABLED);
    this._paused = false;
    this._checker = new StatusChecker(this);
  
    // setup the database
    this._loadDB();
    
    // setup the timer
    this._timer = Cc[TIMER_CONTRACTID].createInstance(Ci.nsITimer);    
    
    // start the update timer
    if (!this.disabled) {
      this._timer.init(this, SERVICE_TIMER, Ci.nsITimer.TYPE_REPEATING_SLACK);
      this._tryCheck();
    }
  },
  
  _unload: function nls__unload() {
    
    // remove the pref observer
    this._branch.QueryInterface(Ci.nsIPrefBranch2).removeObserver("", this);  
    
    // stop the update timer
    this._timer.cancel();
          
    // unload the checker
    this._checker.unload();
  },
  
  _unloadFinal: function nls__unloadFinal() {
         
    // remove observers
    this._obService.removeObserver(this, TOPIC_SHUTDOWN_APP);    
    this._obService.removeObserver(this, TOPIC_SHUTDOWN_XPCOM);
     
    // remove variables
    this._obService = null;    
    this._branch = null;          
    this._disabled = null;
    this._paused = null;
    this._items = null;
    this._conn = null;        
    this._statements = null;    
    this._checker = null;
    this._timer = null; 
  }, 

  _loadDB: function nls__loadDB() {

    // get the database file
    var dir = Cc[DIR_CONTRACTID].getService(Ci.nsIProperties);
    var file = dir.get(SERVICE_DIR, Ci.nsIFile);    
    file.append(SERVICE_FILE);
    
    // open a connection to the database
    var storage = Cc[STORAGE_CONTRACTID].getService(Ci.mozIStorageService);
    try {
      this._conn = storage.openDatabase(file);
    } catch(e) {
      
      // database is corrupt - remove and try again
      if (e.result == 0x805200b) {
        file.remove(false);
        this._conn = storage.openDatabase(file);
        
      // unknown error  
      } else
        throw e;
    }
    
    // new install create tables and version
    if (!this._conn.tableExists("items"))
      this._initDB();
      
    // upgrade the database if not a new install
    else
      this._upgradeDB();
      
    // create the needed statements
    this._statements = {};
    for (name in SERVICE_STMTS)
      this._statements[name] = createStatement(this._conn, SERVICE_STMTS[name]);       
    
    // load the cache
    var statement = this._statements["selectItems"];
    var items = [];
    while (statement.step())
      items[items.length] = createItemFromRow(statement.row);  
    statement.reset();    
    this._items = items;
  },
  
  _initDB: function nls__initDB() {
    for (table in SERVICE_TABLES) {
      try {
        this._conn.executeSimpleSQL("DROP TABLE " + table);
      } catch (e) {}
      this._conn.createTable(table, SERVICE_TABLES[table]);
    }  
    this._conn.executeSimpleSQL("INSERT INTO schema VALUES(" + SERVICE_SCHEMA + ")");
  },
  
  _upgradeDB : function nls__upgradeDB() {
  
    // create the schema table if it does not exist
    var oldVersion = this._getVersion();
    if (!this._conn.tableExists("schema")) {
      this._conn.createTable("schema", SERVICE_TABLES["schema"]);
      this._conn.executeSimpleSQL("INSERT INTO schema VALUES(" + oldVersion + ")");
    }
  
    // upgrade from 0 to 2
    if (oldVersion === 0 && SERVICE_SCHEMA === 2) {
    
      // we drop the feed table and put everything in the items table
      this._conn.beginTransaction();
      try {
      
        // drop the items table
        this._conn.executeSimpleSQL("DROP TABLE items");
      
        // create a new items table
        this._conn.createTable("items", SERVICE_TABLES["items"]);

        // copy feeds to the items table
        var cols = ["id", "parentId", "title", "URL", "iconURL", "refreshRate", 
                    "subscribed", "lastUpdated", "sortIndex"];
        var vals = ["id", "'newslive_root'", "title", "URL", "iconURL", "refreshRate",
                    "subscribed", 0, "sortIndex"];            
        this._conn.executeSimpleSQL("INSERT INTO items(" + cols.join(",") + ") " +
                                    "SELECT " + vals.join(",") + " FROM feeds");      
      
        // drop the feeds table
        this._conn.executeSimpleSQL("DROP TABLE feeds");

        this._conn.commitTransaction();      
      } catch(e) {
        this._conn.rollbackTransaction();
        throw e;
      }
    }
    
    // update the version in the table
    this._conn.executeSimpleSQL("UPDATE schema set version=" + SERVICE_SCHEMA + " where version=" + oldVersion);
  },
  
  _getVersion: function nls__getVersion() {

    var version = 0;
    try {
      // query the schema table
      var select = createStatement(this._conn, "SELECT version FROM schema LIMIT 1");
      if (select.step())
        version = select.row.version;
      select.reset();
        
      // return the version
      return version;
    
    // an error occurred  
    } catch(e) {
      return 0;
    }
  },
    
  _prefUpdate: function nls__prefUpdate(data) {
  
    if (data != SERVICE_DISABLED)
      return;
      
    var wasDisabled = this.disabled;
    var isDisabled = this._branch.getBoolPref(SERVICE_DISABLED);
    this._disabled = isDisabled;
      
      // Went from enabled to disabled stop the timer
    if (isDisabled && !wasDisabled) {
      this._timer.cancel();
      this._notify(null, SERVICE_STATE);
        
      // Went from disabled to enabled start the timer
    } else if (!isDisabled && wasDisabled) {
      this._notify(null, SERVICE_STATE);
      this._timer.init(this, SERVICE_TIMER, Ci.nsITimer.TYPE_REPEATING_SLACK);
      this._tryCheck();
    }
  }, 
  
  _tryCheck: function nls__tryCheck() {
  
    // do nothing if disabled or paused
    if (this.disabled || this.paused)
      return;
      
    // request a check
    this._checker.requestUpdates();
  },
    
  _notify: function nls__notifyObservers(subject, data) {
    try {
      this._obService.notifyObservers(subject, TOPIC_NEWSLIVE, data);
    } catch(e) {
      Components.utils.reportError(e);
    }
  }
};
/*******************************************************************************
 * Generic object used as a factory based on a passed in constructor.
 ******************************************************************************/
function GenericFactory(ctor) {
  this._ctor = ctor;
}  
GenericFactory.prototype = {
  _ctor: null,
  
  // nsISupports
  QueryInterface: function gf_QI(iid) {
    if (iid.equals(Ci.nsIFactory) ||
        iid.equals(Ci.nsISupports))
      return this;
    throw Cr.NS_ERROR_NO_INTERFACE; 
  },
  
  // nsIFactory
  createInstance: function gf_createInstance(outer, iid) {
    if (outer !== null)
      throw Cr.NS_ERROR_NO_AGGREGATION;
    return (new this._ctor()).QueryInterface(iid);
  },

  lockFactory: function gf_lockFactory(loc) {
    throw Cr.NS_ERROR_NOT_IMPLEMENTED;
  }
};
/*******************************************************************************
 * Object implementing the nsIModule interface.  This is used to retrieve 
 * the different components and services the module implements.
 ******************************************************************************/
var Module = {

  // nsISupports
  QueryInterface: function mod_QI(iid) {
    if (iid.equals(Ci.nsIModule) ||
        iid.equals(Ci.nsISupports))
      return this;
    throw Cr.NS_ERROR_NO_INTERFACE;
  },
  
  // nsIModule
  getClassObject: function mod_getClassObject(cm, cid, iid) {
    if (!iid.equals(Ci.nsIFactory))
      throw Cr.NS_ERROR_NOT_IMPLEMENTED;
    
    if (cid.equals(ITEM_CLASSID))
      return new GenericFactory(NewsliveItem);
    else if (cid.equals(SERVICE_CLASSID))
      return new GenericFactory(NewsliveService);
          
    throw Cr.NS_ERROR_NO_INTERFACE;
  },

  registerSelf: function mod_registerSelf(cm, file, location, type) {
    var cr = cm.QueryInterface(Ci.nsIComponentRegistrar);  

    cr.registerFactoryLocation(ITEM_CLASSID, ITEM_CLASSNAME, ITEM_CONTRACTID,
                               file, location, type);
    cr.registerFactoryLocation(SERVICE_CLASSID, SERVICE_CLASSNAME, SERVICE_CONTRACTID,
                               file, location, type);                                                                                           
  },

  unregisterSelf: function mod_unregisterSelf(cm, location, type) {
    var cr = cm.QueryInterface(Ci.nsIComponentRegistrar);
    
    cr.unregisterFactoryLocation(ITEM_CLASSID, location);
    cr.unregisterFactoryLocation(SERVICE_CLASSID, location);
  },

  canUnload: function mod_canUnload(cm) { return true; }
};
/*******************************************************************************
 * Entry point used to retrieve the module object.
 ******************************************************************************/
function NSGetModule(cm, file) { return Module; }